#!/usr/bin/env python

'''Blocked table format.

Usage:
 import blocked_table
 tb = blocked_table.TableBuilder('/tmp/mytable')
 for i in range(1000000):
   tb.add(str(i), 'Value %d' % i)
 del tb

 t = blocked_table.open('/tmp/mytable')
 for k, v in t.iteritems(): print k
 v = t.lookup('this is a key')
'''


from struct import pack, unpack
import bisect
import json
import logging
import os
import re
import zlib

__all__ = [ 'Table', 'TableBuilder', 'dictionary_to_table' ]

class NullCompressor(object):
  @staticmethod
  def compress(str): return str

  @staticmethod
  def decompress(str): return str


COMPRESS_LIB = { 'zlib' : zlib,
                 'none' : NullCompressor }

try:
  import snappy
  COMPRESS_LIB['snappy'] = snappy
except ImportError:
  pass

DEFAULT_BLOCK_SIZE = 32768

class BlockFormat(object):
  MERGED_KEY_VALUE = 1
  SPLIT_KEY_VALUE = 2

try:
  import cjson
  def decode(str):
    return cjson.decode(str)

  def encode(str):
    return cjson.encode(str)

  #logging.info('Using cjson.')
except ImportError:
  import json
  DECODER = json.JSONDecoder()
  ENCODER = json.JSONEncoder(ensure_ascii=True)

  def decode(str):
    return DECODER.decode(str)

  def encode(str):
    return ENCODER.encode(str)

def open_or_create(target):
  os.system('mkdir -p "%s"' % (os.path.dirname(target)))
  return open(target, 'w')

def dictionary_to_table(dict, target):
  '''Convenience method: sort and write 'dict' to table 'target'.'''
  builder = TableBuilder(target)
  for k, v in sorted(dict.items()):
    builder.add(k, v)

def read_int(f):
  packed_val = f.read(4)
  if len(packed_val) != 4:
    return None
  return unpack('I', packed_val)[0]

def read_str(f, decompress=None):
  clen = read_int(f)
  if clen is None:
    return None
  cs = f.read(clen)
  if not decompress:
    return cs
  return decompress(cs)

def skip_str(f):
  clen = read_int(f)
  f.seek(f.tell() + clen)

def write_str(f, s, compress=None):
  if compress:
    cs = compress(s)
  else:
    cs = s
  f.write(pack('I', len(cs)))
  f.write(cs)

class BlockedTableException(Exception):
  pass

class TableIterator(object):
  def __init__(self, data, idx_keys, idx_vals, decompress, block_format):
    self.data = data
    self.idx_keys = idx_keys
    self.idx_vals = idx_vals
    self.decompress = decompress
    self.block_keys = None
    self.block_values = None
    self.block_idx = 0
    self.block_format = block_format
    self.key = None
    self.value = None

    self._next()

  def _next_block(self):
    block_str = read_str(self.data, self.decompress)
    if block_str is None:
      return False
    
    if self.block_format == BlockFormat.MERGED_KEY_VALUE:
      block = decode(block_str)
      self.block_keys = [k for k, v in block]
      self.block_values = [v for k, v in block]
    elif self.block_format == BlockFormat.SPLIT_KEY_VALUE:
      self.block_keys = decode(block_str)
      self.block_values = decode(read_str(self.data, self.decompress))
    else:
      raise NotImplementedException, 'Unknown block format %s' % self.block_format

    self.block_idx = 0
    return True

  def _next(self):
    if not self.block_keys or self.block_idx >= len(self.block_keys):
      if not self._next_block():
        self.key = self.value = None
        return False

    self.key = self.block_keys[self.block_idx]
    self.value = self.block_values[self.block_idx]
    self.block_idx += 1
    return True

  def lookup(self, k):
    if self.seek(k):
      return self.value
    return None

  def seek(self, lk):
    if not self.idx_keys:
      return False

    block_idx = bisect.bisect_left(self.idx_keys, lk)

    if block_idx == len(self.idx_keys) or self.idx_keys[block_idx] >= lk:
      block_idx = max(0, block_idx - 1)

    pos = self.idx_vals[block_idx]
    self.data.seek(pos)
    self._next_block()

    self.block_idx = max(0, bisect.bisect_left(self.block_keys, lk) - 1)
    while self._next():
      if self.key >= lk:
        break

    return self.key == lk

  def __iter__(self):
    if self.key is None:
      raise StopIteration
    while 1:
      yield self.key, self.value
      if not self._next():
        break

    raise StopIteration


class Table(object):
  def __init__(self, f):
    with open(f + '.index', 'rb') as index_file:
      header = decode(read_str(index_file))
      self.compression_format = header['compression']
      self.block_format = header.get('block_format', BlockFormat.MERGED_KEY_VALUE)
      if not self.compression_format in COMPRESS_LIB:
        raise BlockedTableException, ('Compression format %s not available.' %
                                      self.compression_format)
      self.decompress = COMPRESS_LIB[self.compression_format].decompress
      index = decode(read_str(index_file))

    self.idx_keys = [str(k) for k, _ in index]
    self.idx_vals = [v for _, v in index]
    self.name = f

  def iterator(self):
    return TableIterator(open(self.name),
                     self.idx_keys,
                     self.idx_vals,
                     self.decompress,
                     self.block_format)

  def lookup(self, lk):
    return self.iterator().lookup(lk)

  def iteritems(self):
    return self.iterator()


class TableBuilder(object):
  def __init__(self, fname, max_block_size=DEFAULT_BLOCK_SIZE, compression='snappy'):
    if compression is None:
      compression = 'none'

    if not compression in COMPRESS_LIB:
      logging.warn('Compressor %s not available.  Data will not be compressed.' % compression)
      compression = 'none'

    self.compression = COMPRESS_LIB[compression].compress
    self.compression_format = compression

    self.data = open_or_create(fname)
    self.data.seek(0)
    self.index = []
    self.block = []
    self.block_size = 0
    self.max_block_size = max_block_size
    self.name = fname

  def flush(self):
    logging.info('Writing index for table %s', self.data.name)
    self.flush_block()
    with open(self.data.name + '.index', 'w') as index_file:
      write_str(index_file,
                encode({ 'compression' : self.compression_format,
                         'block_format' : BlockFormat.SPLIT_KEY_VALUE  }))
      write_str(index_file, encode(self.index))
    del self.data
    self.data = None

  def __del__(self):
    if self.data:
      self.flush()

  def flush_block(self):
    if not self.block:
      return
    first_key = self.block[0][0]
    #logging.debug('Adding index: (%s, %s)', first_key, self.data.tell())

    self.index.append((first_key, self.data.tell()))
    write_str(self.data, encode([k for k, v in self.block]), self.compression)
    write_str(self.data, encode([v for k, v in self.block]), self.compression)
    self.block = []
    self.block_size = 0

  def add(self, key, value):
    self.block.append((key, value))
    self.block_size += len(encode((key, value)))
    if self.block_size > self.max_block_size:
      self.flush_block()


if __name__ == '__main__':
  import sys
  if len(sys.argv) != 2:
    print >>sys.stderr, 'Usage: %s <table>' % sys.argv[0]
    sys.exit(1)

  bt = Table(sys.argv[1])
  for k, v in bt.iterator():
    print k, v

  
